# coding: utf-8
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .configuration import Config  # noqa
    from ..logging.base import AbstractLogger  # noqa


class AbstractInterface(ABC):

    _config: 'Config'
    _logger: 'AbstractLogger'

    def __init__(
        self, config: 'Config', logger: 'AbstractLogger'
    ):
        self._config = config
        self._logger = logger

    @abstractmethod
    def send(self, request):
        """Отправляет запрос и возвращает результат отправки запроса."""



# Generated by Django 2.2.28 on 2022-07-14 06:50

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('smev_agent_client', '0001_initial'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='entry',
            index=models.Index(fields=['date_time'], name='smev_agent_client_date_ti_e7c85f_idx'),
        ),
    ]



# coding: utf-8

import logging

from .base import AbstractLogger
from .base import Entry


class Logger(AbstractLogger):

    def __init__(self):
        self.__logger = logging.getLogger('smev_agent_client')

    def log(self, entry: Entry) -> None:

        assert isinstance(entry, Entry)

        if entry.error:
            method = self.__logger.error
        else:
            method = self.__logger.info

        method(str(entry))



# coding: utf-8


DEFAULT_REQUEST_TIMEOUT = 300  # Таймаут запроса, секунд
DEFAULT_REQUEST_RETRIES = 5  # Количество повторных попыток
DEFAULT_RETRY_FACTOR = 1  # Шаг увеличения задержки м.д. запросами



# coding: utf-8

from typing import Union

from .configuration import Config


__config: Union[Config, None] = None


def set_config(config: Config):
    """Установка объекта конфигурации пакета."""

    global __config

    assert isinstance(config, Config)

    __config = config


def get_config() -> Config:
    """Получение установленного объекта конфигурации."""

    global __config

    assert isinstance(__config, Config), 'Не произведена настройка клиента'

    return __config



# coding: utf-8


from dataclasses import asdict
from typing import Optional

from ..models import Entry as DBEntry
from .base import AbstractLogger
from .base import Entry


class Logger(AbstractLogger):

    def log(self, entry: Entry) -> Optional[DBEntry]:

        assert isinstance(entry, Entry)

        db_entry = DBEntry.objects.create(**asdict(entry))

        return db_entry



# coding: utf-8
from abc import ABC
from abc import abstractmethod
from dataclasses import asdict
from datetime import datetime
from typing import TYPE_CHECKING
from typing import Optional

from django.utils import timezone
from pydantic.fields import Field


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa
else:
    from pydantic.dataclasses import dataclass  # noqa


@dataclass
class Entry:

    request: str = Field(
        title='Запрос'
    )
    response: Optional[str] = Field(
        title='Ответ на запрос',
        default=None
    )
    error: Optional[str] = Field(
        title='Ошибка',
        default=None
    )
    date_time: datetime = Field(
        default_factory=timezone.now,
        title='Дата и время отправки запроса'
    )

    def __str__(self):
        return (
            '[{date_time}] {request}\n{response}\n{error}'.format(
                **asdict(self)
            )
        )


class AbstractLogger(ABC):

    @abstractmethod
    def log(self, entry) -> None:
        """Логирует сообщение об ошибке."""
        raise NotImplementedError()



# coding: utf-8
from typing import TYPE_CHECKING

from pydantic.fields import Field

from .const import DEFAULT_REQUEST_RETRIES
from .const import DEFAULT_REQUEST_TIMEOUT
from .const import DEFAULT_RETRY_FACTOR


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa
else:
    from pydantic.dataclasses import dataclass  # noqa


@dataclass
class Config:

    """Объект конфигурации пакета."""

    agent_url: str = Field(
        title='Адрес Агента Поставщика данных (schema://host:post)',
        default='http://localhost:8090',
        min_length=1
    )
    system_mnemonics: str = Field(
        title='Мнемоника Агента Поставщика данных',
        min_length=1
    )
    request_retries: int = Field(
        title='Количество повторных попыток',
        default=DEFAULT_REQUEST_RETRIES
    )
    retry_factor: int = Field(
        title='Шаг увеличения задержки м.д. запросами',
        default=DEFAULT_RETRY_FACTOR
    )
    timeout: int = Field(
        title='Таймаут запроса, сек',
        default=DEFAULT_REQUEST_TIMEOUT
    )
    interface: str = 'smev_agent_client.interfaces.rest.OpenAPIInterface'
    logger: str = 'smev_agent_client.logging.db.Logger'



# coding: utf-8

from .base import AbstractInterface
from .rest import OpenAPIInterface
from .rest import OpenAPIRequest
from .rest import Result


__all__ = [
    'AbstractInterface', 'OpenAPIRequest',
    'OpenAPIInterface',
    'Result'
]



# coding: utf-8
from django.db.models.base import Model
from django.db.models.fields import DateTimeField
from django.db.models.fields import TextField
from django.db.models.indexes import Index

from .logging.base import Entry as EntryModel


class Entry(Model):
    """
    Запись журнала взаимодействия с Агентом.
    """

    request = TextField(
        verbose_name=EntryModel.request.title
    )
    response = TextField(
        verbose_name=EntryModel.response.title,
        null=True, blank=True
    )
    error = TextField(
        verbose_name=EntryModel.error.title,
        null=True, blank=True
    )
    date_time = DateTimeField(
        verbose_name=EntryModel.date_time.title
    )

    class Meta:
        verbose_name = verbose_name_plural = 'Журнал запросов'
        indexes = [
            Index(fields=['date_time']),
        ]



# coding: utf-8
from functools import partial
import cgi
import csv
import io
import re


def join_parts(parts, delimiter=';'):
    return delimiter.join(parts)


join_lines = partial(join_parts, delimiter='\n')


def try_decode_csv(value) -> str:
    """Проверка возможности декодирования csv."""

    delimiter_ = ';'

    if issubclass(type(value), bytes):
        value = value.decode('utf-8')

    reader = csv.reader(io.StringIO(value), delimiter=delimiter_)

    result = join_lines(
        join_parts(parts) for parts in reader
    )
    return result


def try_decode_multipart(value):
    boundary = re.compile(
        '(--){1}([a-z1-90]{32})(\\r\\n).*'
    ).match(
        value.decode('utf-8')
    )[2]

    result = cgi.parse_multipart(
        io.BytesIO(value),
        {
            'boundary': boundary.encode('ASCII')
        }
    )
    return result



# coding: utf-8

from typing import TYPE_CHECKING

from .utils import import_string


if TYPE_CHECKING:
    from ..configuration import Config  # noqa
    from .logging.base import AbstractLogger  # noqa
    from .interfaces import AbstractInterface  # noqa


class AgentAdapter:

    _config: 'Config'
    _interface: 'AbstractInterface'
    _logger: 'AbstractLogger'

    def __init__(self):
        self._config = None
        self._load_config()

    def _load_config(self) -> 'Config':
        from smev_agent_client import get_config

        set_config = get_config()

        if self._config is not set_config:
            self._config = set_config

        self._logger = import_string(self._config.logger)()
        self._interface = import_string(
            self._config.interface
        )(
            self._config, self._logger
        )

        return self._config

    def send(self, request):
        """Отправка запроса."""
        return self._interface.send(request)


adapter = AgentAdapter()



# coding: utf-8
from json import load
from pathlib import Path
from urllib.parse import urlparse
from urllib.parse import urlunparse
import logging

from openapi_core.contrib.requests import RequestsOpenAPIRequest
from openapi_core.spec.shortcuts import create_spec
from requests.models import PreparedRequest
from requests.models import Response
from requests.sessions import Session

from smev_agent_client.interfaces.rest import OpenAPIInterface

from .validation import RequestValidator


logger = logging.getLogger('smev_agent_client')


class OpenAPIInterfaceEmulation(OpenAPIInterface):

    def _send_request(
        self, session: Session, http_request: PreparedRequest, **http_kwargs
    ) -> Response:
        """Эмуляция запроса-ответа.

        Получает запрос, вместо отправки валидирует его.
        Код ответа зависит от успешности валидации.
        """

        openapi_request = RequestsOpenAPIRequest(http_request)

        # Из URL удаляется мнемоника агента и basePath спцификации, т.к.
        # это не учитывается в спецификации и поэтому запрос не проходит валидацию
        parsed_urlpattern = urlparse(openapi_request.full_url_pattern)
        scheme, netloc, path, params, query, fragment = parsed_urlpattern
        path = ''.join(f'/{part}' for part in path.split('/')[3:])
        openapi_request.full_url_pattern = urlunparse((
            scheme, netloc, path, params, query, fragment
        ))

        validator = self._get_request_validator()

        result = validator.validate(openapi_request)

        response = Response()

        assert http_request.url
        response.url = http_request.url

        if result.errors:
            logger.error('; '.join(str(err) for err in result.errors))
            response.status_code = 400
        else:
            logger.info('Запрос прошел валидацию')
            response.status_code = 200

        return response

    def _get_request_validator(self):
        with Path(__file__).parent.joinpath('myedu.json').open('r') as spec_file:
            spec_dict = load(spec_file)

        return RequestValidator(create_spec(spec_dict))



# coding: utf-8
import fnmatch

from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.media_types.finders import (
    MediaTypeFinder as BaseMediaTypeFinder)
from openapi_core.validation.request.validators import (
    RequestValidator as BaseRequestValidator)

from . import utils


class MediaTypeFinder(BaseMediaTypeFinder):
    def find(self, request):
        if request.mimetype in self.content:
            return self.content / request.mimetype, request.mimetype

        for key, value in self.content.items():
            if (
                key in request.mimetype or
                fnmatch.fnmatch(request.mimetype, key)
            ):
                return value, key

        raise MediaTypeNotFound(request.mimetype, list(self.content.keys()))


class RequestValidator(BaseRequestValidator):

    media_type_finder_cls = MediaTypeFinder

    def __init__(
        self, *args, custom_media_type_deserializers=None, **kwargs
    ):
        custom_media_type_deserializers: dict = custom_media_type_deserializers or {}
        custom_media_type_deserializers.update({
            'application/octet-stream': utils.try_decode_csv,
            'multipart/form-data': utils.try_decode_multipart
        })
        super().__init__(
            *args, custom_media_type_deserializers=custom_media_type_deserializers
        )

    def _get_media_type(self, content, request_or_response):
        return self.media_type_finder_cls(content).find(request_or_response)



# coding: utf-8
from abc import ABC
from abc import abstractmethod
from typing import IO
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib.parse import urljoin
import traceback

from pydantic.fields import Field
from pydantic.main import BaseModel
from requests.adapters import HTTPAdapter
from requests.adapters import Retry
from requests.exceptions import RequestException
from requests.models import PreparedRequest
from requests.models import Request as HttpRequest
from requests.models import Response
from requests.sessions import Session

from smev_agent_client.const import DEFAULT_REQUEST_TIMEOUT
from smev_agent_client.logging.base import Entry

from .base import AbstractInterface


class AbstractRESTRequest(ABC):

    """Абстрактный REST-запрос."""


class SQLRequest(AbstractRESTRequest):
    content_type: str = 'application/x-www-form-urlencoded; charset=utf-8'
    accept_version: str = '1'

    async_request: bool = False

    sql: str
    priority: str = 'NORMAL'  # Literal['NORMAL', 'HIGH']
    timeout: Optional[str] = None
    params: list = Field(default_factory=list)
    maxrows: Optional[str] = None


class OpenAPIRequest(AbstractRESTRequest):

    @abstractmethod
    def get_url(self) -> str:
        """URL запроса."""

    @abstractmethod
    def get_method(self) -> str:
        """Метод запроса.

        'get', 'post', 'delete', 'patch', 'put' or 'options'.
        """

    def get_params(self) -> dict:
        """Параметры Query string (get)."""
        return {}

    def get_headers(self) -> dict:
        """Заголовки запроса."""
        return {}

    def get_data(self):
        """Тело запроса."""
        return None

    def get_files(self) -> List[Tuple['str', IO]]:
        """Файлы в запросе.

        (
            ('files', Path(name1).open('rb'),
            ('files', Path(name2).open('rb'),
        )
        """
        return []

    def get_timeout(self) -> Union[int, None]:
        """Таймаут запроса, сек."""
        return None


class Result(BaseModel):

    class Config:
        arbitrary_types_allowed = True

    request: AbstractRESTRequest
    response: Optional[Any] = None
    http_request: Optional[Any] = None
    error: Optional[str] = None
    log: Optional[Any] = None


class SQLInterface(AbstractInterface):

    """Режим SQL-запросов.

    В синхронном режиме получение результата осуществляется путем выполнения
    HTTP-запроса от ИС Потребителя к Агенту ПОДД.

    В рамках HTTP-запроса (метод POST) передается SQL-запрос, в ответе возвращается
    результат выполнения SQL-запроса.
    """

    @abstractmethod
    def send(self, request: SQLRequest) -> Result:
        """Отправляет запрос и возвращает результат отправки запроса."""
        raise NotImplementedError()


class OpenAPIInterface(AbstractInterface):

    """Синхронный режим по спецификации OpenAPI.

    Все запросы выполняются в синхронном режиме в соответствии с загруженной
    в ПОДД СМЭВ спецификацией OpenAPI.
-
    Получение результата осуществляется путем выполнения HTTP-запроса от
    ИС Потребителя данных к Агенту Поставщика данных.
    """

    def send(self, request: OpenAPIRequest) -> Result:
        session = Session()

        http_adapter = HTTPAdapter(
            max_retries=Retry(
                total=self._config.request_retries,
                backoff_factor=self._config.retry_factor
            )
        )

        session.mount('https://', http_adapter)
        session.mount('http://', http_adapter)

        args = (
            request.get_method().upper(),
            self._determine_request_url(request)
        )
        kwargs = dict(
            headers=request.get_headers(),
            params=request.get_params(),
            data=request.get_data(),
            files=request.get_files(),
        )

        response = error = proxies = stream = verify = cert = None

        http_request = self._prepare_http_request(session, *args, **kwargs)

        http_kwargs = dict(
            timeout=self._select_timeout(request)
        )
        http_kwargs.update(session.merge_environment_settings(
            http_request.url, proxies, stream, verify, cert
        ))

        try:
            response = self._send_request(session, http_request, **http_kwargs)
            response.raise_for_status()
        except RequestException as err:
            error = self._format_error(err)

        entry = Entry(
            request=self._format_http_request(http_request),
            response=self._format_response(response),
            error=error,
        )

        log = self._log(entry)

        return Result(
            request=request,
            http_request=http_request,
            response=response,
            error=error,
            log=log,
        )

    def _log(self, entry: Entry):
        result = self._logger.log(entry)

        return result

    def _send_request(
        self, session: Session, http_request: PreparedRequest, **http_kwargs
    ) -> Response:
        return session.send(http_request, **http_kwargs)

    def _determine_request_url(self, request):
        """Определение URL запроса.

        URL = схема + адрес адаптера + мнемоника адаптера + path запроса.
        """
        return urljoin(
            self._config.agent_url,
            f'/{self._config.system_mnemonics}{request.get_url()}'
        )

    def _select_timeout(self, request):
        """Определение таймаута.

        Приоритет таймаутов:
            - из запроса
            - из настройки клиента
            - стандартный
        """
        return (
            request.get_timeout() or
            self._config.timeout or
            DEFAULT_REQUEST_TIMEOUT
        )

    def _prepare_http_request(
        self, session, *args, **kwargs
    ) -> PreparedRequest:
        return session.prepare_request(HttpRequest(
            *args, **kwargs
        ))

    def _format_http_request(
        self, http_request: PreparedRequest
    ) -> str:
        body = (
            http_request.body.decode('utf-8') if http_request.body else ''
        )
        return (
            f"[{http_request.method}] {http_request.url}\n\n"
            f"{http_request.headers}\n\n"
            f"{body}"
        )

    def _format_error(self, error: Union[Exception, None]) -> str:
        if not error:
            return ''
        return f'{str(error)}\n\n{traceback.format_exc()}'

    def _format_response(self, response: Union[Response, None]) -> str:
        if response is None:
            return ''
        return f'[{response.status_code}] {response.url}\n\n{response.headers}\n\n{response.text}'



# Generated by Django 2.2.28 on 2022-07-08 07:36

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('request', models.TextField(verbose_name='Запрос')),
                ('response', models.TextField(blank=True, null=True, verbose_name='Ответ на запрос')),
                ('error', models.TextField(blank=True, null=True, verbose_name='Ошибка')),
                ('date_time', models.DateTimeField(verbose_name='Дата и время отправки запроса')),
            ],
            options={
                'verbose_name': 'Журнал запросов',
                'verbose_name_plural': 'Журнал запросов',
            },
        ),
    ]
